/*
 * CpConnection.java
 *
 * Copyright  Paul Pinault, Disk91 2003.
 *
 * Tout dtenteur d'une copie de ce programme et des fichiers de 
 * documentation associs (le  logiciel ) bnficie des droits 
 * d'exploitation, de copie, de modification, de rutilisation, de 
 * publication, de distribution gratuite ou commerciale, et reste libre 
 * de distribuer l'ensemble et/ou d'en vendre des copies en octroyant ces 
 * mmes droits aux utilisateurs sous rserve de signaler son utilisation 
 *  son auteur par mail, indiquant votre identit et le nom de votre 
 * projet,  l'adresse suivante : dpl@disk91.com, ce signalement implique 
 * de votre part un accord d'utilisation de ces informations dans le but de 
 * promouvoir le dit logiciel. En outre, les droit sont accords sous 
 * rserve d'inclure le copyright ci-dessus et l'avertissement suivant dans 
 * toute copie, totale ou partielle, du programme :
 *
 * CE PROGRAMME EST FOURNI  TEL QU'EN L'TAT , SANS AUCUNE GARANTIE DE 
 * QUELQUE NATURE, IMPLICITE OU EXPLICITE, QUANT  SON UTILISATION 
 * COMMERCIALE, PROFESSIONNELLE, LGALE OU NON, OU AUTRE. L'AUTEUR NE PEUT 
 * EN AUCUN CAS TRE TENU POUR RESPONSABLE DE QUELQUE DOMMAGE OU PRJUDICE 
 * DIRECT, INDIRECT, SECONDAIRE OU ACCESSOIRE DCOULANT DE L'UTILISATION DE 
 * TOUT OU PARTIE DE CE PROGRAMME.
 *
 * Les noms Paul Pianult et  Disk91 ne saurait tre utilis dans le but 
 * de promouvoir ou de lgitimer un produit driv de ce programme sans 
 * autorisation crite pralable  cette effet.
 *
 * Created on 4 juillet 2003, 15:59
 */

package disk.ConnexionPool;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Date;
import org.apache.log4j.*;

/**
 * Classe CpConnection
 * Cette classe encapsule un objet connexion et ajoute des informations
 * quand a sa date de cration, sa derniere date d'utilisation.
 * 
 * La classe gere ces timeout, elle ne ralise aucun traitement si la
 * dure de vie est dpassen mais elle est capable de calculer et 
 * indiquer ces dpassements sur demande... 
 *
 * @author  Paul Pinault
 * @version 1.0
 * @date    2003/07/04
 */
 public class CpConnection {

    /** Constante utilise pour le Timeout de la connexion : dure de vie en sec */
    protected static final int TIMEOUT_CNX = (30*60);
    /** Constante utilise pour le Timeout de la connexion : dure max d'utilisation en s */
    protected static final int TIMEOUT_REQ = 15;
    
    //                      Logger Log4j
    
    protected static Category log = Category.getInstance(CpConnection.class.getName()); 
    
    //                      Quelques donnes utiles
    
    /** La connexion encapsule */
    protected Connection    cnx;
    /** Un statement encapsul pour retrouver la bonne connexion a la fin */
    protected Statement     sta;
    /** Date de creation de la connexion */
    protected long          creation;
    /** Date de derniere utilisation de la connexion */
    protected long          utilisation;
    /** Indique que la Connexion est en cours d'utilisation */
    private   boolean       enCours;
    /** nom de l'utilisateur (pour memoire) */
    protected String        user;
     

    //                     Gestion des TimeOut
    
    /** Timeout a partir de la cration de la connexion */
    private   int           timeoutCnx;
    /** Timeout durant un requete */
    private   int           timeoutReq;
    /** Indique que le timeout Req a deja ete not pour eviter d'engorger les logs */
    private   boolean       treqLog;
    
    //               Calcul du temps d'utilisation
    
    /** Temps total d'utilisation en ms */
    private   long          totalUseTime;
    /** Nombre de mesures */
    private   int           nbUseTime;
    
    /** taux d'utilisation, date de dpart de la mesure */
    private   long          tuStartTime;
    /** taux d'utilisation, temps total d'utilisation depuis dpart */
    private   long          tuUseTime;

    // ====================== CONSTRUCTION =========================    
    
   /** 
    * Construction d'une CpConnection a partir d'un objet
    * Connection existant.
    * Ajoute l'heure de cration et d'utilisation
    * Les timeout sont fixe a partir de valeurs par defaut.
    */
    public CpConnection(Connection cnx) {
        init(cnx, CpConnection.TIMEOUT_CNX, CpConnection.TIMEOUT_REQ);
    }

   /** 
    * Construction d'une CpConnection a partir d'un objet
    * Connection existant.
    * Ce constructeur permet de fixer en outre les parametres
    * de dure de vie.
    *
    * @param toCnx  Dure de vie de la connexion en sec -1 si infinie
    * @param toReq  Dure de vie d'une requete en sec -1 si indinie
    * @param cnx    La connexion
    */
    public CpConnection(Connection cnx,int toCnx, int toReq) {
        init(cnx, toCnx, toReq);
    }

   /**
    * Fonction d'init commune aux diffrents construsteurs ....
    */ 
    private void init(Connection cnx, int toCnx, int toReq ) {
        // Mise a jour param
        this.cnx          = cnx;
        this.creation     = System.currentTimeMillis();
        this.utilisation  = this.creation;
        this.enCours      = false;
        this.timeoutCnx   = toCnx;
        this.timeoutReq   = toReq;
        // Init des compteurs pour stats
        this.totalUseTime = 0;
        this.nbUseTime    = 0;
        this.resetUsingStat();
        // trace ...
        log.debug( "Creation Connexion ("+this.cnx+") a ["+new Date(this.utilisation)+"] "
                 +"timeoutCnx("+this.timeoutCnx+"s) timeoutReq("+this.timeoutReq+"s) ");
    }

    // ======================== Fermeture ==========================    
    
    /** 
     * Ferme la connexion en cours et indique la conenxion comme
     * prise pour qu'elle ne puisse plus etre rclame...
     */
     public void close() {
         log.debug("Fermeture de la Connexion ("+this.cnx+")");
         this.enCours = true;       // lock 
         try {
             this.cnx.close();
             this.cnx = null;
         } catch (SQLException e){ 
             log.error(" Connexion ("+this.cnx+") probleme lors de la fermeture : "+e.toString());
         }
     }
        
    
    // =========================== OUTILS ========================== 
    
   /**
    * Recupration de l'objet Connexion encapsul, on en profite
    * pour faire une mise a jour de son heure d'utilisation
    *
    * @param    user    chaine permettant d'identifier l'utilisateur de la
    *                   connexion
    */
    public Connection getConnection(String user) {
        if ( this.enCours == false ) {
           this.utilisation = System.currentTimeMillis();
           this.enCours     = true;
           this.treqLog     = false;
           this.user = user;
           log.debug( "\""+user+"\" prend la connexion("+this.cnx+") a ["+new Date(this.utilisation)+"] ");
           return this.cnx;
        }
        log.error("La connexion("+this.cnx+") n'est pas libre ");
        return null;
    }

    /**
     * Interroge pour savoir si il s'agit de la meme Connection 
     * ou non
     * @param   cnx     Connexion a comparer
     * @return  true/false si identique/ou non.
     */
     public boolean isSameConnection(Connection cnx) {
         if ( cnx == this.cnx )
             return true;
         else
             return false;
     }
    
   /**
    * Retour de l'objet connexion, sert juste a la mise a l'heure
    * une nouvelle fois ... pour indiquer sa fin d'utilisation
    */ 
    public void forgetConnection() {
        if ( this.enCours == true ) {
           long time          = System.currentTimeMillis();
           // mise a jour des infos pour le calcul de temps moyen
           // d'utilisation de la connexion
           this.totalUseTime += (time - this.utilisation);
           this.nbUseTime++;
           this.tuUseTime    += (time - this.utilisation);
           // mise a jour de la derniere date d'utilisation
           this.utilisation   = time;
           this.enCours       = false;
           log.debug( "Libere la connexion("+this.cnx+") a ["+new Date(this.utilisation)+"] ");
        } else {
           log.warn("Demande de liberation de la connexion("+this.cnx+") dj libre ");
        }
    }

    /**
     * Plutot qu'une connexion on peut demander un Statement, dans ce
     * cas il sera mmoris pour le retrouver par la suite, a chaque
     * demande, un nouvel objet Statement est cre.
     *
     * @param    user    chaine permettant d'identifier l'utilisateur de la
     *                   connexion
     */
     public Statement getStatement(String user) throws ConnexionPoolException {
          log.debug("Creation d'un Statement");
          try {
              this.getConnection(user);
              sta = cnx.createStatement();
              if (sta == null)
                  log.error("Statement null ...");
          } catch (SQLException e) {
              log.error("Exception sql :"+e.toString());
              throw new ConnexionPoolException("Exception sql : "+e.toString());
          }
          return sta; 
     }
     
    /**
     * Interroge pour savoir si il s'agit du Statement de cet objet ou
     * non
     * @param   sta     Statement a comparer
     * @return  true/false si identique/ou non.
     */
     public boolean isSameStatement(Statement sta) {
         if ( sta == this.sta )
             return true;
         else
             return false;
     }
     
    /**
     * Libere le statement prcdemment allou
     */
     public void forgetStatement() {
         log.debug("Liberation Statement");
         if ( sta == null ){
             log.warn("Liberation d'un statement non utilise");
             return;
         }
         try {
             sta.close();
             this.forgetConnection();
         } catch (SQLException e) {
             log.error("Exception sql :"+e.toString());
         }
         sta = null;
     }
    
    // ======================== GESTION DES TIMEOUTS ====================
    
    /**
     * indique si la connexion n'est pas en fin de vie
     * @return true si elle doit etre dtruite
     */
     public boolean isTooOld() {
         if ( this.timeoutCnx == -1 ) return false;
         long now  = System.currentTimeMillis();
         long life = now - this.creation;
         if ( life > (this.timeoutCnx*1000) ) {
             log.debug("La connexion("+this.cnx+") est trop vieille");
             return true;
         } else 
             return false;
     }
    
    /**
     * indique si la connexion n'est pas en cours d'utilisation depuis
     * trop longtemps.
     *
     * @return true si elle doit etre dtruite par scurit ....
     */
     public boolean isBusy() {
         /* pour eviter en engorgement des traces ... */
         if ( this.treqLog == true ) return false;
         if ( this.timeoutReq == -1 ) return false;
         long now  = System.currentTimeMillis();
         long life = now - this.utilisation;
         if ( life > (this.timeoutReq*1000) ) {
             this.treqLog = true;
             log.warn("La connexion("+this.cnx+") est utilise depuis trop longtemps");
             return true;
         } else 
             return false;
     }
     
    /**
     * Retourne le temps moyen d'utilisation de la connexion, ce temps
     * peut etre utilis pour mettre  jour le timeoutReq de faon
     * dynamique. Toutfois, il ne faudra pas oublier d'ajouter une marge
     * de scurit...
     *
     * @return un temps moyen d'utilisation en ms
     */
     public long meanUseTime() {
         long mut = this.totalUseTime / this.nbUseTime;
         log.debug("La connexion("+this.cnx+") rpond en moyenne en "+mut+"ms ");
         return mut;
     }
     
    /**
     * Modification dynamique de l'indicateur de TimeoutRequest utilis 
     * pour savoir si la connexion est Busy (ou non rendue)...
     * 
     * @param toReq     nouveau timeoutReq a utiliser exprim en seconde
     */
     public void setTimeoutReq(int toReq) {
         log.debug("Modification timeout requete de la connexion("+this.cnx+") : "+toReq+"s ");
         this.timeoutReq = toReq;
     }

    /**
     * Remise a zero des outils de mesure de taux d'utlisation
     */
     public void resetUsingStat() {
        this.tuStartTime  = System.currentTimeMillis();
        this.tuUseTime    = 0;
     }
         
       
    /**
     * Retourne le taux moyen d'utilisation d'une connexion
     * ==> son temps d'utilisation / son temps d'existance
     *
     * @return un entier exprimant le pourcentage d'utilisation de la
     *  connexion.
     */
     public int getUsingStat() {
         long now  = System.currentTimeMillis();
         int v = (int)((this.tuUseTime / (float)(now - this.tuStartTime))*100);
         log.debug("La connexion("+this.cnx+") a un taux d'occupation de "+v+"% ");
         return v;
     }

    /**
     * Retourne le taux moyen d'utilisation d'une connexion depuis sa creation
     * ==> son temps d'utilisation / son temps d'existance
     *
     * @return un entier exprimant le pourcentage d'utilisation de la
     *  connexion.
     */
     public int getTotalUsingStat() {
         long now  = System.currentTimeMillis();
         int v = (int)((this.totalUseTime / (float)(now - this.creation))*100);
         return v;
     }
     
    /**
     * Retourne la chaine user 
     */
     public String getUser() {
         return this.user;
     }

    /**
     * Date du dernier acces
     */
     public java.util.Date getLastAccessTime() {
         return new java.util.Date(this.utilisation);
     }
 }
