/*
 * Pool.java
 *
 * Copyright  Paul Pinault, Disk91 2003.
 *
 * Tout dtenteur d'une copie de ce programme et des fichiers de 
 * documentation associs (le  logiciel ) bnficie des droits 
 * d'exploitation, de copie, de modification, de rutilisation, de 
 * publication, de distribution gratuite ou commerciale, et reste libre 
 * de distribuer l'ensemble et/ou d'en vendre des copies en octroyant ces 
 * mmes droits aux utilisateurs sous rserve de signaler son utilisation 
 *  son auteur par mail, indiquant votre identit et le nom de votre 
 * projet,  l'adresse suivante : dpl@disk91.com, ce signalement implique 
 * de votre part un accord d'utilisation de ces informations dans le but de 
 * promouvoir le dit logiciel. En outre, les droit sont accords sous 
 * rserve d'inclure le copyright ci-dessus et l'avertissement suivant dans 
 * toute copie, totale ou partielle, du programme :
 *
 * CE PROGRAMME EST FOURNI  TEL QU'EN L'TAT , SANS AUCUNE GARANTIE DE 
 * QUELQUE NATURE, IMPLICITE OU EXPLICITE, QUANT  SON UTILISATION 
 * COMMERCIALE, PROFESSIONNELLE, LGALE OU NON, OU AUTRE. L'AUTEUR NE PEUT 
 * EN AUCUN CAS TRE TENU POUR RESPONSABLE DE QUELQUE DOMMAGE OU PRJUDICE 
 * DIRECT, INDIRECT, SECONDAIRE OU ACCESSOIRE DCOULANT DE L'UTILISATION DE 
 * TOUT OU PARTIE DE CE PROGRAMME.
 *
 * Les noms Paul Pianult et  Disk91 ne saurait tre utilis dans le but 
 * de promouvoir ou de lgitimer un produit driv de ce programme sans 
 * autorisation crite pralable  cette effet.
 *
 * Created on 4 juillet 2003, 15:16
 */

package disk.ConnexionPool;

import disk.tools.sllist.SLList;
import java.util.Iterator;
import java.sql.*;
import org.apache.log4j.*;

/**
 * Classe ConnexionPool : Abstract
 * Classe modle pour un pool de connexion.
 * C'est a l'utilisateur de driver cette classe pour raliser la
 * connexion a la base de donne et grer le singleton, de cette
 * faon, il est possible d'avoir simultanement plusieurs pool
 * lis a diffrentes bases ou diffrents drivers ... 
 * 
 * @author  Paul Pinault
 * @version 1.1 
 * @date    2003/07/04
 */
 public abstract class Pool {
   
    // ==================== PARAMETRES ==============================
     
    /** Parametre par defaut : taille Min du POOL */
    private static final int    POOL_MIN_LEN = 2;
    /** Parametre par defaut : taille Max du POOL */
    private static final int    POOL_MAX_LEN = 10;
    /** Parametre par defaut : increment d'allocation du POOL */
    private static final int    POOL_INC     = 2;

    // ===================== ELEMENTS DU POOL =======================
  
    /** Liste des Connexions libres pouvant etre prises */
    protected SLList   freeList;
    /** Liste des Connexion en cours d'utilisation */
    protected SLList   useList;
    
    /** taille minimale du pool */
    protected int          poolMinLen;
    /** taille maximale du pool */
    protected int          poolMaxLen;
    /** increment dans le pool */
    protected int          poolIncrement;
    
    /** informations permettant la connexion a la base */
    protected String        driver;
    protected String        path;
    protected String        user;
    protected String        pass;
    
    /** logger log4j */
    protected static Category log = Category.getInstance(Pool.class.getName());
    
    /** Thread de gestion */
    protected Thread        poolThread;

    /** Semaphore pour l'acces a la liste ; visibilit package */
    Object          lock;
    
    // ===================== ELEMENTS STATISTIQUE =======================
    /** Demandes de statement et/ou connexions */
    private   int           demandes;
    /** Echec d'allocation de Statement et/ou connexions */
    private   int           echec;
    
    
    
    // ======================= CONSTRUCTION =========================
     
    /** 
     * Le constructeur ne peut etre appel que depuis une classe fille
     * le Pool en lui meme n'existe pas, il est abstrait
     *
     * @param driver nom de la classe correspondant au driver pour 
     *               chargement
     * @param path   chemin de la base de donne de la forme 
     *               "jdbc:mysql://localhost ..."
     * @param user    nom de l'utilisateur pour la connexion a la base
     * @param pass    mot de passe pour la connexion a cette base
     */
     protected Pool(String driver, String path, String user, String pass ) {
        initPool(driver, path, user, pass, 
                 Pool.POOL_MIN_LEN, Pool.POOL_MAX_LEN, Pool.POOL_INC);
     }
     
    /**
     * Constrcution d'un pool avec mes parametres de taille min, max
     * et increment
     *
     * @param driver nom de la classe correspondant au driver pour 
     *               chargement
     * @param path   chemin de la base de donne de la forme 
     *               "jdbc:mysql://localhost ..."
     * @param user   nom de l'utilisateur pour la connexion a la base
     * @param pass   mot de passe pour la connexion a cette base
     * @param pMin   Taille minimale du Pool en nombre de connexions ouvertes
     * @param pMax   Taille maximale du Pool en nombre de connexions ouvertes -1 = inf
     * @param pInc   Nombre de connexion a ouvrir a chaque augmentation
     */
     protected Pool(String driver, String path, String user, String pass, 
                    int pMin, int pMax, int pInc) {
         initPool(driver, path, user, pass, pMin, pMax, pInc);
     }
     
    /**
     * Fonction d'initialisation rendant commun le code des diffrents
     * constructeurs utilisables
     */
     private void initPool(String driver, String path, String user, String pass,
                           int pMin, int pMax, int pInc) {

         // Objet le plus simple pour semaphore
         lock       = new int[1];       

         // Init des parametres
         this.driver        = driver;
         this.path          = path;
         this.user          = user;
         this.pass          = pass;
         this.poolMinLen    = pMin;
         this.poolMaxLen    = pMax;
         this.poolIncrement = pInc;
         
         // Traces
         log.warn("Cration d'un pool pour la base : "+this.path);

         // Init des listes
         this.freeList = new SLList();
         this.useList  = new SLList();
         
         // Init des premieres connexions
         try {
            this.addConnexion(this.poolMinLen);
         } catch (ConnexionPoolException e) {
             log.error("Ajout de connexion au pool avec erreur : "+e.toString());
         }
         
         // Lancement du Thread de gestion
         poolThread = new Thread( new PoolThread(this));
         poolThread.start();
         
         // Init stat
         this.demandes = 0;
         this.echec    = 0;
     }

    // ======================== DESTRUCTION =========================
     
     /**
      * Fermeture du pool, on peut liberer toutes les connexions existantes
      */
      protected void close() {
         synchronized ( this.lock ) { 
            log.warn("Fermeture du pool pour la base "+this.path);
            this.poolThread.destroy();

            Iterator it = this.freeList.iterator();
            while (it.hasNext()) {
                 CpConnection cnx = (CpConnection)it.next();
                 cnx.close();
            }
            for ( int i = 0 ; i < this.freeList.size() ; i++ ) 
                 this.freeList.remove(i);

            it = this.useList.iterator();
            while (it.hasNext()) {
                 CpConnection cnx = (CpConnection)it.next();
                 cnx.close();
            }
            for ( int i = 0 ; i < this.useList.size() ; i++ ) 
                 this.useList.remove(i);
         }
      }
         
      protected void finalize() {
          close();
      }
          
     
     
    // ================= AJOUT/SUP ELEMENTS DANS LE POOL ================
    
     /**
      * Creation de nouvelles connexions a jouter dans le pool
      * par dfaut elles seront bien sure libre ...
      * Cette fonction peut engendrer une ConnexionPoolException
      * /!\ Necessite une synchro externe
      *
      * @param nb       indique le nombre de connexion a ajouter
      *                 dans le pool
      * @return         false s'il y a eu un probleme (par exemple
      *                 nombre de connexion MAX atteind)
      */
      private boolean addConnexion( int nb ) throws ConnexionPoolException {
          // verifie que l'on ne depasse pas le max...
          int max;
          max = this.poolMaxLen - (this.freeList.size() + this.useList.size()); 
          if ( max == 0 ) return false;
          if ( nb > max ) nb = max;
          
          log.debug("Ajout de "+nb+" Connexion dans le pool ...");
          try {
            Class.forName(this.driver).newInstance();
            for ( int i = 0 ; i < nb ; i++ ) {
               CpConnection cnx = new CpConnection(DriverManager.getConnection(this.path,this.user, this.pass));
               this.freeList.addLast(cnx);
            }
          } catch (ClassNotFoundException e) {
              throw new ConnexionPoolException("Driver invalide, non trouv ");
          } catch (InstantiationException e) {
              throw new ConnexionPoolException("Instanciation impossible ");
          } catch (IllegalAccessException e) {
              throw new ConnexionPoolException("Accs illgal ");
          } catch ( SQLException e) {
              throw new ConnexionPoolException("Exception sql : "+e.toString());
          }
          return true;
      }

      /**
       * Version synchronisee de addConnexion avec visibilit package
       */
       boolean addConnexionS( int nb ) throws ConnexionPoolException {
           synchronized ( this.lock ) {
              return this.addConnexion(nb);
           }
       }
          
      
     /**
      * Suppression de connexions dans le pool, on ne pourra en
      * supprimer que dans la freeList et on devra veiller a ce
      * qu'il en reste au moins le nb minimum ...
      * On enleve les moins utilises a la fin normalement pour
      * ne pas gener les stats d'utilisation...
      * /!\ Necessite une synchro externe
      *
      * @param nb       indique le nombre de connexion a retirer
      *                 dans le pool
      * @return         false s'il y a eu un probleme (par exemple
      *                 nombre de connexion MAX atteind)
      */
      private boolean delConnexion( int nb ) {
          log.debug("Suppression de "+nb+" connexions");
          // nombre de connexions en cours...
          int tot;
          int free;

          free = this.freeList.size();
          tot  = free + this.useList.size(); 
              
          // plus de connexions a supprimer
          if ( tot <= this.poolMinLen )
              return false;
          // nb est trop grand
          if ( (tot - nb) <= this.poolMinLen )
              nb = tot - this.poolMinLen;
          // on se limite aux elements de la freeList
          if ( nb > free ) nb = free;
              
          log.debug("Suppression effective de "+nb+" connexions");
          for ( int i = 0 ; i < nb ; i++) {
              CpConnection cpc = (CpConnection)this.freeList.removeLast();
              cpc.close();
          }
          return true;
      }
      
      /**
       * Version synchronisee de delConnexion avec visibilit package
       */
       boolean delConnexionS( int nb ) {
           synchronized ( this.lock ) {
              return this.delConnexion(nb);
           }
       }
      
    // =================== DEMANDE D'UN ELEMENT ===================
    
     /**
      * Cette fonction permet d'obtenir une Connexion du Pool, la connexion
      * sera mise dans la useList, la connexion de tete est prise.
      * S'il n'y a pas de connexion libre, on va en allouer de nouvelle
      * par pas d'increment, toutefois le mcanisme d'allocation doit etre
      * anticip par le thread de surveillance, mais au cas o, ici on 
      * fait l'allocation .. 
      *
      * @return Un objet CpConnection.
      */
      protected CpConnection getFreeConnection() throws ConnexionPoolException {
          synchronized ( this.lock ) {
             this.demandes++;
             log.debug("Recherche d'une connexion libre ...");
             CpConnection cpc; 
             if ( this.freeList.isEmpty() == true ) {
                  // Il manque des connexion on va en allouer de nouvelles
                  this.addConnexion(Pool.POOL_INC);
             }
             if ( this.freeList.isEmpty() == false ) {
                  // cas simple il reste des connexions
                  cpc = (CpConnection)this.freeList.removeFirst();
             } else {
                  this.echec++;
                  log.error("Impossible de trouver une Connexion libre");
                  throw new ConnexionPoolException("Pas de connexion disponible");
             }
             // On ajoute l'objet pris a la liste de objet utilise
             // en fin, car on estime qu'ils sont libr dans le meme ordre
             this.useList.addLast(cpc);
             return cpc;
          }
      }
 
     /**
      * Methode pour l'acces a l'objet Connection a partir de l'objet
      * CpConnection. Appele principalement getFreeConnection()
      *
      * @param  user    Chaine identifiant l'utilisateur de la connexion
      * @return Un objet de type Connection
      */
      public Connection getConnection(String user) throws ConnexionPoolException {
          log.debug("Distribution d'une Connexion...");
          return this.getFreeConnection().getConnection(user);
      }
      
      
     /**
      * Fonction retournant un statement, puisque souvant c'est simplement
      * cel qu'il nous faut, cette fonction fait simplement appel 
      * getFreeConnection et extrait un statement de la connexion.
      *
      * @param  user    Chaine identifiant l'utilisateur de la connexion
      * @return Un objet Statement.
      */
      public Statement getStatement(String user) throws ConnexionPoolException {
          log.debug("Distribution d'un Statement...");
          return this.getFreeConnection().getStatement(user);
      }
              
    // =================== RESTITUTION D'UN ELEMENT ===================
          
     /**
      * Fonction de restitution d'une connexion, on va rechercher dans
      * la liste des connexion celle qui correspond ...
      *
      * @param  cnx     La connexion a liberer
      */
      public void forgetConnection( Connection cnx ) {
          synchronized ( this.lock ) {
             log.debug("Restitution de la connexion "+cnx);
             CpConnection cpc = null;
             boolean      find = false;

             Iterator it = this.useList.iterator();
             while ( it.hasNext() && find == false ) {
                cpc = (CpConnection)it.next();
                if ( cpc.isSameConnection(cnx) == true ) {
                    this.useList.remove(cpc);
                    find = true;
                }
             }
             if ( cpc != null && find == true ) {
                 cpc.forgetConnection();
                 this.freeList.addFirst(cpc);
             } else {
                 log.error("Connexion non trouve dans ce pool");
             }
          }
      }
      
     /**
      * Fonction de restitution d'un statement, on va rechercher dans
      * la liste des connexion celle qui correspond ...
      *
      * @param  sta     Le statement a liberer
      */
      public void forgetStatement( Statement sta ) {
          synchronized ( this.lock ) {
             log.debug("Restitution du Statement "+sta);
             CpConnection cpc = null;
             boolean      find = false;
             Iterator it = this.useList.iterator();
             while ( it.hasNext() && find == false ) {
                cpc = (CpConnection)it.next();
                if ( cpc.isSameStatement(sta) == true ) {
                    this.useList.remove(cpc);
                    find = true;
                }
             }
             if ( cpc != null && find == true ) {
                 cpc.forgetStatement();
                 this.freeList.addFirst(cpc);
             } else {
                 log.error("Statement non trouv dans ce pool");
             }
          }
      }

     // ============================ GETTER =============================
      
      
      /** Getter for property freeList.
       * visibilit : package 
       * @return Value of property freeList.
       */
      disk.tools.sllist.SLList getFreeList() {
          return freeList;
      } 
      
      
      /** Getter for property useList.
       * visibilit : package 
       * @return Value of property useList.
       */
      disk.tools.sllist.SLList getUseList() {
          return useList;
      }
      
      /** Increment du pool
       * @return valeur du pool increment
       */
      public int getPoolIncrement() {
          return this.poolIncrement;
      }
      
      
     /** Nombre des demandes de connexions
      * @return le nombre de demandes de connexions
      */
      public int getPoolRequest() {
          return this.demandes;
      }
      
     /** Nombre d'echec d'obtention de Connexion
      * @return le nombre des echec
      */
      public int getPoolRequestError() {
          return this.echec;
      }  
    /**
     * @return Pool's database path
     * 
     * @author paul
     * @version 1.0
     */
    public String getPath() {
        return path;
    }

 }
