/* 
 * PoolThread.java
 *  
 * Copyright  Paul Pinault, Disk91 2003.
 *
 * Tout dtenteur d'une copie de ce programme et des fichiers de 
 * documentation associs (le  logiciel ) bnficie des droits 
 * d'exploitation, de copie, de modification, de rutilisation, de 
 * publication, de distribution gratuite ou commerciale, et reste libre 
 * de distribuer l'ensemble et/ou d'en vendre des copies en octroyant ces 
 * mmes droits aux utilisateurs sous rserve de signaler son utilisation 
 *  son auteur par mail, indiquant votre identit et le nom de votre 
 * projet,  l'adresse suivante : dpl@disk91.com, ce signalement implique 
 * de votre part un accord d'utilisation de ces informations dans le but de 
 * promouvoir le dit logiciel. En outre, les droit sont accords sous 
 * rserve d'inclure le copyright ci-dessus et l'avertissement suivant dans 
 * toute copie, totale ou partielle, du programme :
 *
 * CE PROGRAMME EST FOURNI  TEL QU'EN L'TAT , SANS AUCUNE GARANTIE DE 
 * QUELQUE NATURE, IMPLICITE OU EXPLICITE, QUANT  SON UTILISATION 
 * COMMERCIALE, PROFESSIONNELLE, LGALE OU NON, OU AUTRE. L'AUTEUR NE PEUT 
 * EN AUCUN CAS TRE TENU POUR RESPONSABLE DE QUELQUE DOMMAGE OU PRJUDICE 
 * DIRECT, INDIRECT, SECONDAIRE OU ACCESSOIRE DCOULANT DE L'UTILISATION DE 
 * TOUT OU PARTIE DE CE PROGRAMME.
 *
 * Les noms Paul Pianult et  Disk91 ne saurait tre utilis dans le but 
 * de promouvoir ou de lgitimer un produit driv de ce programme sans 
 * autorisation crite pralable  cette effet.
 *
 * Created on 7 juillet 2003, 12:27
 */

package disk.ConnexionPool;
import java.util.Iterator;
import java.util.Date;
import org.apache.log4j.*;

/**
 * Classe PoolThread.
 * Thread de surveillance et gestion du Pool, il detruit automatiquement
 * les connexions rop vieille ou bloque, redimensionne le pool en 
 * fonction des besoins.
 *
 * @author  Paul Pinault
 * @version 1.0
 * @date    2003/07/07
 */
public class PoolThread implements Runnable {
    
    /** temps de someil du Thread en ms */
    protected static final int  SLEEP_TIME = 1000;
    /** temps pour le resizing synamique en ms */
    protected static final int  DYNRESIZ_TIME = (10*60*1000);
    /** temps pour l'affichage de stats du pool en ms */
    protected static final int  STAT_TIME = (6*60*60*1000);
    
    /** Le pool auquel se refere le thread */
    protected Pool    pool;

    /** logger log4j */
    protected static Category log = Category.getInstance(PoolThread.class.getName());

    /** Compteur de passage pour le dynResizing */
    private int dynCounter;
    /** Compteur de passage pour les stats */
    private int statCounter;
    
    /** Creates a new instance of PoolThread */
    public PoolThread(Pool pool) {
        log.debug("Creation du Thread");
        this.pool = pool;
    }

    
   /**
    * Methode principale du thread, elle verifie la duree de vie des
    * objets dans le pool, elle vrifie la charge des objets presents
    * en enleve ou en ajoute au besoin ...
    */
    public void run() {
        log.debug("Demarrage du Thread");
        this.dynCounter = 0;
        while ( true ) {
            log.debug("Execution Thread");
            // Surveille les connexion bloques car non rendues ...
            this.busyFix();
            // Replace les connexions trop anciennes
            this.deadFix();
            // vide le pool et anticipe les besoins a venir
            this.dynResize();
            // affiche l'tat du pool rgulierement
            this.statPool();
            
            try {
               Thread.sleep(PoolThread.SLEEP_TIME);
            } catch(InterruptedException e) { }
        }
    }
    
  // ================= LIBERATION DES CONNEXIONS NON RENDUES ==========
    
   /** 
    * Surveille les connexions en cours d'utilisation, si elles le 
    * sont depuis trop longtemps, elles sont busy (non rendues par
    * l'utilisateur), ceci provient normalement d'un bug dans le
    * code utilisateur d'o l'emission d'un warning continu jusqu'a
    * correction...
    */
    private void busyFix() { 
       Iterator it = this.pool.getUseList().iterator();
       while ( it.hasNext() ) {
         CpConnection cpc = (CpConnection)it.next();
         if ( cpc.isBusy() == true ) {
            log.warn("La connexion dtenue par \""+cpc.getUser()+"\" n'a pas t libre");
         }
      }
    }

   /**
    * Surveille les connexions et supprime celles qui sont en fin de vie
    * seul le pool des connexions libre est scrut ....
    */
    private void deadFix() {
        int closed = 0;
        Iterator it = this.pool.getFreeList().iterator();
        while ( it.hasNext() ) {
             CpConnection cpc = (CpConnection)it.next();
             if ( cpc.isTooOld() == true ) {
                 log.debug("Connexion "+cpc+" devenue trop vielle");
                 it.remove();
                 cpc.close();
                 closed++;
             }
        }
        // creation de connexions pour remplacer celle perdues
        if (closed > 0) {
            try {
               this.pool.addConnexionS(closed);
            } catch ( ConnexionPoolException e ) {
                log.warn("Exception lors de l'ajout d'elements : "+e.toString());
            }
        }
    }
    
  // ==================== GESTION DYNAMIQUE DE LA TAILLE ================ 
    
    /**
     * regulierement on va regarder le taux d'utilisation de chaque
     * Connexion, en dessous de 20% on supprime la connexion
     * Au dessus de 60% on va allouer de nouvelles connexions en 
     * prvision
     */
     private void dynResize() {
        
        if ( this.dynCounter > (PoolThread.DYNRESIZ_TIME/PoolThread.SLEEP_TIME) ) {
            this.dynCounter = 0;
            
            int meanUse = 0;
            int nbUse   = 0;
            
            Iterator it = this.pool.getFreeList().iterator();
            while ( it.hasNext() ) {
               CpConnection cpc = (CpConnection)it.next();
               meanUse += cpc.getUsingStat();
               nbUse++;
               cpc.resetUsingStat();
            }
            it = this.pool.getUseList().iterator();
            while ( it.hasNext() ) {
               CpConnection cpc = (CpConnection)it.next();
               meanUse += cpc.getUsingStat();
               nbUse++;
               cpc.resetUsingStat();
            }
            if ( nbUse > 0 ) {
                meanUse = meanUse / nbUse;
                log.debug("Taux d'utilisation moyen actuel : "+meanUse);
                if ( meanUse < 20 ) {
                    this.pool.delConnexionS(this.pool.getPoolIncrement());
                } else if (meanUse > 60) {
                    try {
                       this.pool.addConnexionS(this.pool.getPoolIncrement());
                    } catch (ConnexionPoolException e) {
                       log.error("Probleme d'allocation de nouvelles connexions");
                    }
                }
            }
        }
        else this.dynCounter++;
     }

  // =========================== AFFICHAGE STATS ========================= 
     
    /**
     * Affiche l'tat du pool de facon rguliere si la tempo associe
     * est positive
     */
     private void statPool() {

        if ( PoolThread.STAT_TIME < 0 ) return;
         
        if ( this.statCounter > (PoolThread.STAT_TIME/PoolThread.SLEEP_TIME) ) {
            Iterator it;
            int      i;
            
            this.statCounter = 0;
            System.out.println("+ Etat du pool ["+this.pool.getPath()+"] le "+new Date(System.currentTimeMillis()) );
            System.out.println("   Nombre des demandes de Connexion : "+this.pool.getPoolRequest());
            System.out.println("   Nombre des echec : "+this.pool.getPoolRequestError());
            
            System.out.println("   Elements disponibles ("+this.pool.getFreeList().size()+") :");
            it = this.pool.getFreeList().iterator();
            i  = 1;
            while ( it.hasNext() ) {
               CpConnection cpc = (CpConnection)it.next();
               System.out.println("      - "+i+". Utilisation ["+cpc.getTotalUsingStat()+"%] Dernier acces ["+cpc.getLastAccessTime()+"]");
               i++;
            }

            System.out.println("   Elements occups ("+this.pool.getUseList().size()+") :");
            it = this.pool.getUseList().iterator();
            i  = 1;
            while ( it.hasNext() ) {
               CpConnection cpc = (CpConnection)it.next();
               System.out.println("      - "+i+". Utilisation ["+cpc.getTotalUsingStat()+"%] Dernier acces ["+cpc.getLastAccessTime()+"]");
               i++;
            }
        }
        else this.statCounter++;
     }
     
}
