/*
 * SLList.java
 *
 * Copyright  Paul Pinault, Disk91 2003.
 *
 * Tout dtenteur d'une copie de ce programme et des fichiers de 
 * documentation associs (le  logiciel ) bnficie des droits 
 * d'exploitation, de copie, de modification, de rutilisation, de 
 * publication, de distribution gratuite ou commerciale, et reste libre 
 * de distribuer l'ensemble et/ou d'en vendre des copies en octroyant ces 
 * mmes droits aux utilisateurs sous rserve de signaler son utilisation 
 *  son auteur par mail, indiquant votre identit et le nom de votre 
 * projet,  l'adresse suivante : dpl@disk91.com, ce signalement implique 
 * de votre part un accord d'utilisation de ces informations dans le but de 
 * promouvoir le dit logiciel. En outre, les droit sont accords sous 
 * rserve d'inclure le copyright ci-dessus et l'avertissement suivant dans 
 * toute copie, totale ou partielle, du programme :
 *
 * CE PROGRAMME EST FOURNI  TEL QU'EN L'TAT , SANS AUCUNE GARANTIE DE 
 * QUELQUE NATURE, IMPLICITE OU EXPLICITE, QUANT  SON UTILISATION 
 * COMMERCIALE, PROFESSIONNELLE, LGALE OU NON, OU AUTRE. L'AUTEUR NE PEUT 
 * EN AUCUN CAS TRE TENU POUR RESPONSABLE DE QUELQUE DOMMAGE OU PRJUDICE 
 * DIRECT, INDIRECT, SECONDAIRE OU ACCESSOIRE DCOULANT DE L'UTILISATION DE 
 * TOUT OU PARTIE DE CE PROGRAMME.
 *
 * Les noms Paul Pianult et  Disk91 ne saurait tre utilis dans le but 
 * de promouvoir ou de lgitimer un produit driv de ce programme sans 
 * autorisation crite pralable  cette effet.
 *
 * Created on 10 juillet 2003, 11:25
 */

package disk.tools.sllist;

/**
 * Classe Synchronized Linked List SLList
 * Cette classe est une liste chainee synchronise non base sur un Vector
 * Synchronise, elle est donc plus simple d'emploie que la LinkedList et
 * sans doute plus rapide puisque rellement base sur une liste chainee.
 * Elle implemente une grosse partie des methodes d'une List, toutefois
 * celles concernant les index et Collections ne sont ralises et 
 * retourne des exceptions... 
 *
 * @author  Paul Pinault
 * @version 1.0
 * @date    2003/07/10
 */
public class SLList implements java.lang.Cloneable, java.util.List {
    
    /** Element de tete, visibilit package */
    SLListElement   head;
    /** Element de queue */
    private SLListElement   queue;
    /** Taille de la liste */
    private int             len;
    /** Semaphore pour l'acces a la liste ; visibilit package */
    Object          lock;
    
 // ========================= CONSTRUCTION / DESTRUCTION ======================   
    
    /** 
     * Constructeur : cration de la liste
     * @version 1.0
     */
     public SLList() {
         this.head  = null;
         this.queue = null;
         this.len   = 0;
         lock       = new int[1];       // Objet le plus simple pour semaphore
     }

    /**
     * Constructeur par recopie : clone
     * Sans iterateur car la protection ne serait pas suffisante ...
     */
     public Object clone() {
         SLList s = new SLList();
         synchronized( this.lock ) {
             SLListElement c = this.head;
             while ( c != null ) {
                 Object n = c.getElement();
                 c = c.getNext();
                 s.addLast(n);
             }
         }
         return s;
     }
     

 // ========================= AJOUT D'ELEMENTS ================================     
     
    /**
     * Ajout d'un element en fin de liste
     *
     * @param o - l'lement a ajouter en fin de liste
     */
     public void addLast(Object o) {
         synchronized(this.lock) {
            SLListElement e =  new SLListElement(o);
            if (this.len >0) {
                this.queue.setNext(e);
                e.setPrec(this.queue);
            } else {
                this.head = e;
            }
            this.queue = e;
            this.len++;
         }
     }
     
    /**
     * Ajout d'un element en fin de liste
     *
     * @param o - l'lement a ajouter en fin de liste
     * @retrun true, toujours, compatibilit LinkedList
     */
     public boolean add(Object o) {
         this.addLast(o);
         return true;
     }

    /**
     * Ajout d'un lment en debut de liste
     * 
     * @param o - l'lment a ajouter en debut de liste
     */
     public void addFirst(Object o) {
         synchronized(this.lock) {
            SLListElement e =  new SLListElement(o);
            if (this.len >0) {
                e.setNext(this.head);
                this.head.setPrec(e);
            } else {
                this.queue = e;
            }
            this.head = e;
            this.len++;
         }
     }
     
    /**
     * Ajout d'un lment devant un autre
     * 
     * Visibilit : package
     *
     * /!\ Doit etre synchronize exterieurement /!\
     *
     * @param e - L'lment de rfrence
     * @param o - L'objet a ajouter
     */
     void addBefore(SLListElement e, Object o) {
         SLListElement n =  new SLListElement(o);
         if ( e == this.head ) {
            if (this.len > 0) {
                n.setNext(this.head);
                this.head.setPrec(n);
            } else {
                this.queue = n;
            }
            this.head = n;
         } else if( e != null ) {
            e.getPrec().setNext(n);
            n.setPrec(e.getPrec());

            n.setNext(e);
            e.setPrec(n);
         } else return;
         this.len++;
     }
             
     
         
 // ========================= SUPPRESSION D'ELEMENTS ==========================     
     
    /**
     * Suppression de tous les lments de la liste
     */
     public void clear() {
         synchronized(this.lock) {
             SLListElement c =  this.head;
             SLListElement s;
             while (c != null) {
                 s = c.getNext();
                 c.clear();
                 c = s;
             }
             this.head = this.queue = null;
             this.len  = 0;
         }
     }
     
    /**
     * Suppression d'un lment en particulier dans la liste
     * visibilit package seulement 
     *
     * /!\ Doit etre synchronize exterieurement /!\
     *
     */
     void remove( SLListElement e ) {
         if ( e != null ) {
             if ( e != this.head ) 
                e.getPrec().setNext(e.getNext());
             else
                this.head = e.getNext();
                
             if ( e != this.queue )
                e.getNext().setPrec(e.getPrec());
             else
                this.queue = e.getPrec();
             
             this.len--;
         }
     }
     
    /**
     * Suppression d'un lment dans la liste.
     *
     * Recherche la premiere occurance de l'objet pass en parametre et
     * le retire de la liste.
     *
     * @param o - L'lment a supprimer
     * @return true si la liste contenait l'lment a dtruire
     */
     public boolean remove(Object o) {
         synchronized ( this.lock ) {
             SLListElement c = this.head;
             while ( c != null ) {
                 Object n = c.getElement();
                 if (    (n == null && o == null)
                      || (n != null && n.equals(o)) ) {
                    this.remove(c);
                    return true;
                 }
                 c = c.getNext();
             }
             return false;
         }
     }

    /**
     * Suppression du dernier lment de la liste
     *
     * Supprime le dernier lment de la liste
     *
     * @return l'objet supprim
     * @exception java.util.NoSuchElementException si la liste est vide
     */
     public Object removeLast() throws java.util.NoSuchElementException {
         synchronized ( this.lock ) {
             if ( this.len > 0 ) {
                SLListElement c = this.queue;
                Object        n = c.getElement();
                this.remove(c);
                return n;
             } else throw new java.util.NoSuchElementException();
         }
     }
     
    /**
     * Suppression du premier lment de la liste
     *
     * Supprime le premier lment de la liste
     *
     * @return l'objet supprim
     * @exception java.util.NoSuchElementException si la liste est vide
     */
     public Object removeFirst() throws java.util.NoSuchElementException {
         synchronized ( this.lock ) {
             if ( this.len > 0 ) {
                SLListElement c = this.head;
                Object        n = c.getElement();
                this.remove(c);
                return n;
             } else throw new java.util.NoSuchElementException();
         }
     }
     

 // ========================= ACCESS AUX ELEMENTS =============================     

    /**
     * Retourne le premier lement
     *
     * @return l'objet inclus dans l'lment de tte
     * @exception java.util.NoSuchElementException si la liste est vide
     */
     public Object getFirst() throws java.util.NoSuchElementException {
         if ( this.len == 0 ) throw new java.util.NoSuchElementException();
         synchronized( this.lock ) {
             return this.head.getElement();
         }
     }
         
    /**
     * Retourne le dernier lement
     *
     * @return l'objet inclus dans l'lment de queue
     * @exception java.util.NoSuchElementException si la liste est vide
     */
     public Object getLast() throws java.util.NoSuchElementException {
         if ( this.len == 0 ) throw new java.util.NoSuchElementException();
         synchronized( this.lock ) {
             return this.queue.getElement();
         }
     }
     
 // ========================= RECHERCHE D'ELEMENTS ============================     
     
    /**
     * Retroune true si la liste contient l'lment indiqu
     *
     * Plus formellement, retourne true si et seulement si la liste contient
     * au moins un lment identique a o.
     *
     * @param o - L'lment dont la prsence est a tester
     * @return true si l'lment est prsent dans la liste
     */
     public boolean contains(Object o) {
         synchronized ( this.lock ) {
             SLListElement c = this.head;
             while ( c != null ) {
                 Object n = c.getElement();
                 if (    (n == null && o == null)
                      || (n != null && n.equals(o)) ) {
                    return true;
                 }
                 c = c.getNext();
             }
             return false;
         }
     }

 // ========================= ITERATIONS ======================================     

    /**
     * Construit et retourne un itrateur sur la liste
     *
     * @return un objet de type itrator
     */
     public java.util.Iterator iterator() {
         synchronized ( this.lock ) {
            return new SLListIterator(this);
         }
     }
     
    /**
     * Construit et retourne un itrateur de liste sur la liste
     * seule la valeur est accepte pour index : les index ne
     * sont pas supports.
     *
     * @return un objet de type listIterator
     */
     public java.util.ListIterator listIterator(int index) throws java.lang.IllegalArgumentException {
         if ( index != 0 ) throw new java.lang.IllegalArgumentException();
         synchronized ( this.lock ) {
            return new SLListIterator(this);
         }
     }

     public java.util.ListIterator listIterator() {
         synchronized ( this.lock ) {
            return new SLListIterator(this);
         }
     }

     
 // ========================= DIVERS ==========================================     
     
    /**
     * Retourne le nombre d'lement dans la liste
     * 
     * @return le nombre d'lments dans la liste
     */
     public int size() {
         synchronized( this.lock ) {
            return this.len;
         }
     }
     
    /**
     * Retourne true si la liste est vide
     *
     * @return true si la liste est vide.
     */
     public boolean isEmpty() {
         synchronized( this.lock ) {
            return ( this.len == 0 );
         }
     }
     
    /**
     * Calcule une clef de hashage en fonction de la liste et son
     * contenu.
     *
     * @return clef de hashage.
     */
     public int hashCode() {
         int hashCode = 1;
         synchronized ( this.lock ) {
             SLListElement c = this.head;
             while ( c != null ) {
                 Object n = c.getElement();
                 hashCode = 31*hashCode + (n==null ? 0 : n.hashCode());
                 c = c.getNext();
             }
         }
         return hashCode;
     }

 // ========================= METHODES NON SUPPORTEES ======================   
     
    public void add(int index, Object element) throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }

    public boolean addAll(java.util.Collection c) throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }
    
    public boolean addAll(int index, java.util.Collection c) throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }
    
    public boolean containsAll( java.util.Collection c ) throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }
    
    public Object get( int index ) throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }
    
    public int indexOf( Object o ) throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }
    
    public int lastIndexOf( Object o ) throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }
    
    public Object remove( int index ) throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }
    
    public boolean removeAll( java.util.Collection c ) throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }
    
    public boolean retainAll( java.util.Collection c ) throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }
    
    public Object set( int index, Object o ) throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }
    
    public java.util.List subList( int fromIndex, int toIndex) throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }
    
    public Object[] toArray() throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }
    
    public Object[] toArray(Object a[]) throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }
    
    
}
