/*
 * SLListIterator.java
 *
 * Copyright  Paul Pinault, Disk91 2003.
 *
 * Tout dtenteur d'une copie de ce programme et des fichiers de 
 * documentation associs (le  logiciel ) bnficie des droits 
 * d'exploitation, de copie, de modification, de rutilisation, de 
 * publication, de distribution gratuite ou commerciale, et reste libre 
 * de distribuer l'ensemble et/ou d'en vendre des copies en octroyant ces 
 * mmes droits aux utilisateurs sous rserve de signaler son utilisation 
 *  son auteur par mail, indiquant votre identit et le nom de votre 
 * projet,  l'adresse suivante : dpl@disk91.com, ce signalement implique 
 * de votre part un accord d'utilisation de ces informations dans le but de 
 * promouvoir le dit logiciel. En outre, les droit sont accords sous 
 * rserve d'inclure le copyright ci-dessus et l'avertissement suivant dans 
 * toute copie, totale ou partielle, du programme :
 *
 * CE PROGRAMME EST FOURNI  TEL QU'EN L'TAT , SANS AUCUNE GARANTIE DE 
 * QUELQUE NATURE, IMPLICITE OU EXPLICITE, QUANT  SON UTILISATION 
 * COMMERCIALE, PROFESSIONNELLE, LGALE OU NON, OU AUTRE. L'AUTEUR NE PEUT 
 * EN AUCUN CAS TRE TENU POUR RESPONSABLE DE QUELQUE DOMMAGE OU PRJUDICE 
 * DIRECT, INDIRECT, SECONDAIRE OU ACCESSOIRE DCOULANT DE L'UTILISATION DE 
 * TOUT OU PARTIE DE CE PROGRAMME.
 *
 * Les noms Paul Pianult et  Disk91 ne saurait tre utilis dans le but 
 * de promouvoir ou de lgitimer un produit driv de ce programme sans 
 * autorisation crite pralable  cette effet.
 *
 * Created on 10 juillet 2003, 12:41
 */

package disk.tools.sllist;

/**
 * Iterateur sur une Synchronized Linked List
 *
 * L'Iterator est autant que possible synchronis sur la liste mais
 * cette synchro ne peut pas etre sure a 100%, entre deux acces consecutifs
 * a l'iterateur, la liste peut avoir t modifie et l'Iterateur peut
 * etre devenu invalide ...
 *
 * @author  Paul Pinault
 * @version 1.0
 * @date    2003/07/10
 */
public class SLListIterator implements java.util.Iterator, java.util.ListIterator {
    
    /** Pointeur vers le suivant dans la liste */
    private SLListElement   next;
    /** Pointeur vers l'lement en cours d'utilisation pour remove */
    private SLListElement   cur;
    /** Pointeur vers l'lement prcdant celui en cours dans la liste */
    private SLListElement   prev;

    /** indique que l'lement en cours d'utilisation a t supprim */
    private boolean         deleted;
    /** Pointeur vers la liste   */
    private SLList          liste;

    
   /** 
    * Creation d'un nouvel itrateur a partir d'une liste
    *
    * @param liste - la liste sur laquelle on itere
    * @param head  - le premier lment de la liste
    */
    public SLListIterator(SLList liste) {
        this.liste   = liste;
        this.next    = liste.head;
        this.cur     = null;
        this.prev    = null;
        this.deleted = false;
    }
 
   /**
    * Retourne true si l'itrateur a d'autre lments.
    *
    * Plus prcisemment, indique si l'lment courant est valide ou
    * nul...
    *
    * @return true s'il y a d'autres lements
    */
    public boolean hasNext() {
        synchronized ( this.liste.lock ) {
           return ( this.next != null );
        }
    }
    
   /**
    * Retourne true si l'itrateur a un lment prcdant
    *
    * @return true s'il y a un lment prcdant
    */
    public boolean hasPrevious() {
        synchronized ( this.liste.lock ) {
           return ( this.prev != null );
        }
    }
    
   /**
    * Retourne le contenu de l'lment courant et pointe le suivant
    *
    * @return le contenu de l'lement courant.
    * @exception java.util.NoSuchElementException si il n'y a plus de 
    *            suivant.
    */ 
    public Object next() throws java.util.NoSuchElementException {
        synchronized ( this.liste.lock ) {
           if ( this.next != null ) {
               this.prev = this.cur;
               this.cur  = this.next;
               this.next = this.next.getNext();
	       this.deleted = false;
               return this.cur.getElement();
           }
           else
               throw new java.util.NoSuchElementException();
        }
    }
    
   /**
    * Retourne le contenu de l'lment prcdant et pointe le suivant sur celui
    * qui tait le courant
    *
    * @return le contenu de l'lment prcdant
    * @exception java.util.NoSuchElementException s'il n'y a plus de prcdant
    */
    public Object previous() throws java.util.NoSuchElementException {
        synchronized ( this.liste.lock ) {
           if ( this.prev != null ) {
               this.next = this.cur;
               this.cur  = this.prev;
               this.prev = this.prev.getPrec();
	       this.deleted = false;
               return this.cur.getElement();
           }
           else
               throw new java.util.NoSuchElementException();
        }
    } 
       
            
   /**
    * Supprime le dernier lment retourn par getNext de la liste
    * pour cel, on fait appel a la liste ...
    *
    * @exception java.lang.IllegalStateException si il n'y a pas d'lment
    *            en cours ou que l'objet a dja t supprim.
    */
    public void remove() throws java.lang.IllegalStateException {
        synchronized ( this.liste.lock ) {

           if ( this.deleted == true || this.cur == null ) 
              throw new java.lang.IllegalStateException();
        
           this.liste.remove(this.cur);
           this.deleted = true;
        }
    }
    
   /**
    * Insere un lement dans la liste
    *
    * Ajoute un lment apres l'lment en cours et avant l'lment a venir
    * ceci n'affecte donc pas la suite de l'itration sauf si l'on fait
    * appel a previous...
    *
    * @param o - l'objet a ajouter
    */
    public void add(Object o) {
        synchronized ( this.liste.lock ) {
           this.liste.addBefore(this.cur, o);   
           if ( this.cur == null ) {
              this.prev = this.liste.head;
           } else {
              this.prev = this.cur.getPrec();
           }
        }
    }
     
   /**
    * Remplace l'lment courant par l'objet pass en paramtre
    *
    * @param o L'objet servant de remplacant
    * @exception java.lang.IllegalStateException si l'on a pas d'lment courant
    */
    public void set(Object o) throws java.lang.IllegalStateException {
        if ( this.cur == null ) throw new java.lang.IllegalStateException();
        this.cur.setElement(o);
    }
    
    
   /**
    * Fonction nextIndex et previous index : non implment car sans 
    * aucun sans dans le cas d'un vrai liste chainee
    * Au besoin, utiliser une LinkedList a la place ...
    * @return UnsupportedOperationException
    * @exception java.lang.UnsupportedOperationException toujours
    */
    public int nextIndex() throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }

   /**
    * Fonction nextIndex et previous index : non implment car sans 
    * aucun sans dans le cas d'un vrai liste chainee
    * Au besoin, utiliser une LinkedList a la place ...
    * @return UnsupportedOperationException
    * @exception java.lang.UnsupportedOperationException toujours
    */
    public int previousIndex() throws java.lang.UnsupportedOperationException {
        throw new java.lang.UnsupportedOperationException();
    }
    
}
